import torch
import unittest
import torchbearer
from mock import Mock

import torchbearer.variational.auto_encoder as ae


class SimpleAE(ae.AutoEncoderBase):
    def encode(self, x, state=None):
        return x, state

    def decode(self, sample, state=None):
        return sample, state


class TestAutoEncoder(unittest.TestCase):
    def test_empty_methods(self):
        base = ae.AutoEncoderBase(10)

        self.assertRaises(NotImplementedError, lambda: base.encode(0))
        self.assertRaises(NotImplementedError, lambda: base.decode(0))

    def test_forward_call_counts(self):
        latents = 10
        x = torch.rand(1)
        state = {torchbearer.X: x, torchbearer.Y_TRUE: 1}

        model = SimpleAE(latents)
        model.encode = Mock()
        model.decode = Mock()

        model.forward(x, state)
        self.assertTrue(model.encode.call_count == 1)
        self.assertTrue(model.decode.call_count == 1)

        model.forward(x, state)
        self.assertTrue(model.encode.call_count == 2)
        self.assertTrue(model.decode.call_count == 2)

    def test_forward_call_args(self):
        latents = 10
        x = torch.rand(1)
        state = {torchbearer.X: x, torchbearer.Y_TRUE: 1}

        model = SimpleAE(latents)
        model.encode = Mock()
        model.encode.return_value = None
        model.decode = Mock()

        model.forward(x, state)
        self.assertTrue(model.encode.call_args[0] == (x, state))
        self.assertTrue(model.decode.call_args[0] == (None, state))

    def test_forward_replace_y_true(self):
        latents = 10
        x = torch.rand(1)
        state = {torchbearer.X: x, torchbearer.Y_TRUE: 1}

        model = SimpleAE(latents)
        model.encode = Mock()
        model.encode.return_value = None
        model.decode = Mock()

        model.forward(x, state)
        self.assertTrue(model.decode.call_args[0][1][torchbearer.Y_TRUE] == x)

    def test_latent_dims(self):
        latents = 10
        x = torch.rand(1)
        state = {torchbearer.X: x, torchbearer.Y_TRUE: 1}

        model = SimpleAE(latents)
        self.assertTrue(model.latent_dims == latents)


import unittest
from mock import Mock

import torch

import torchbearer
from torchbearer.variational import DivergenceBase, SimpleNormalUnitNormalKL, SimpleNormalSimpleNormalKL, SimpleNormal, SimpleWeibull, SimpleWeibullSimpleWeibullKL, SimpleExponentialSimpleExponentialKL, SimpleExponential

key = torchbearer.state_key('divergence_test')


class TestDivergenceBase(unittest.TestCase):
    def test_on_criterion(self):
        divergence = DivergenceBase({'test': key}).with_sum_sum_reduction()
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 4)
        self.assertTrue(state[torchbearer.LOSS].requires_grad)
        divergence.compute.assert_called_once_with(test=1)

    def test_on_criterion_validation(self):
        divergence = DivergenceBase({'test': key}).with_sum_sum_reduction()
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion_validation(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 4)
        divergence.compute.assert_called_once_with(test=1)

    def test_empty_compute(self):
        divergence = DivergenceBase({'test': key})
        self.assertRaises(NotImplementedError, divergence.compute)

    def test_custom_post_fcn(self):
        divergence = DivergenceBase({'test': key}).with_sum_sum_reduction().with_post_function(lambda loss: loss * 2)
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion_validation(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 8)
        divergence.compute.assert_called_once_with(test=1)

    def test_store(self):
        divergence = DivergenceBase({'test': key}, state_key=key).with_sum_sum_reduction().with_post_function(lambda loss: loss * 2)
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion_validation(state)
        self.assertTrue(state[key].item() == 4)
        self.assertTrue(state[torchbearer.LOSS].item() == 8)
        divergence.compute.assert_called_once_with(test=1)

    def test_reductions(self):
        divergence = DivergenceBase({'test': key}).with_sum_sum_reduction()
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion_validation(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 4)

        divergence = DivergenceBase({'test': key}).with_sum_mean_reduction()
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_criterion_validation(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 2)

    def test_with_beta(self):
        divergence = DivergenceBase({'test': key}).with_sum_sum_reduction().with_beta(4)
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion_validation(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 16)
        divergence.compute.assert_called_once_with(test=1)

    def test_with_linear_capacity(self):
        divergence = DivergenceBase({'test': key}).with_sum_sum_reduction().with_linear_capacity(min_c=0, max_c=6, steps=6, gamma=2)
        divergence.compute = Mock(return_value=torch.ones((2, 2), requires_grad=True))

        state = {
            torchbearer.LOSS: torch.zeros(1, requires_grad=True),
            key: 1
        }

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 8)
        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_step_training(state)

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 6)
        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_step_training(state)

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 4)
        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_step_training(state)

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 2)
        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_step_training(state)

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 0)
        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_step_training(state)

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 2)
        state[torchbearer.LOSS] = torch.zeros(1)
        divergence.on_step_training(state)

        divergence.on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS].item() == 4)


class TestSimpleNormalUnitNormalKL(unittest.TestCase):
    def test_divergence(self):
        callback = SimpleNormalUnitNormalKL(input_key=key)
        dist = SimpleNormal(torch.zeros(2, 2), torch.ones(2, 2) * -1.3863)
        self.assertTrue(((callback.compute(dist) - 0.3181).abs() < 0.0001).all())


class TestSimpleNormalSimpleNormalKL(unittest.TestCase):
    def test_divergence(self):
        callback = SimpleNormalSimpleNormalKL(key, key)
        input = SimpleNormal(torch.zeros(2, 2), torch.ones(2, 2) * -1.3863)
        target = SimpleNormal(torch.ones(2, 2), torch.ones(2, 2) * 1.3863)
        self.assertTrue(((callback.compute(input, target) - 1.0425).abs() < 0.0001).all())

class TestSimpleWeibullSimpleWeibullKL(unittest.TestCase):
    def test_divergence(self):
        callback = SimpleWeibullSimpleWeibullKL(key, key)
        input = SimpleWeibull(torch.ones(2, 2), torch.zeros(2, 2) + 0.5)
        target = SimpleWeibull(torch.ones(2, 2), torch.ones(2, 2) * 5)
        self.assertTrue(((callback.compute(input, target) - 3628803.7500).abs() < 0.0001).all())


class TestSimpleExponentialSimpleExponentialKL(unittest.TestCase):
    def test_divergence(self):
        callback = SimpleExponentialSimpleExponentialKL(key, key)
        input = SimpleExponential(torch.ones(2, 2) - 1.6931)
        target = SimpleExponential(torch.zeros(2, 2))
        self.assertTrue(((callback.compute(input, target) - 0.3068).abs() < 0.0001).all())


import math
import unittest
from mock import patch, Mock

import torch

from torchbearer.variational import SimpleDistribution, SimpleNormal, SimpleUniform, SimpleExponential, SimpleWeibull

class TestEmptyMethods(unittest.TestCase):
    def test_methods(self):
        dist = SimpleDistribution()

        self.assertTrue(dist.support is None)
        self.assertTrue(dist.arg_constraints is None)
        self.assertTrue(dist.mean is None)
        self.assertTrue(dist.variance is None)

        self.assertTrue(dist.expand(torch.Size()) is None)
        self.assertTrue(dist.cdf(1) is None)
        self.assertTrue(dist.icdf(1) is None)
        self.assertTrue(dist.enumerate_support() is None)
        self.assertTrue(dist.entropy() is None)

        self.assertRaises(NotImplementedError, lambda: dist.rsample())
        self.assertRaises(NotImplementedError, lambda: dist.log_prob(1))


class TestSimpleNormal(unittest.TestCase):
    @patch('torchbearer.variational.distributions.torch.normal')
    def test_rsample_tensor(self, normal):
        mu = torch.ones(2, 2)
        logvar = torch.zeros(2, 2)

        dist = SimpleNormal(mu, logvar)

        normal.side_effect = lambda mu, std: mu
        self.assertTrue(dist.rsample(sample_shape=torch.Size([2])).sum() == 8)

        normal.side_effect = lambda mu, std: std
        self.assertTrue(dist.rsample(sample_shape=torch.Size([2])).sum() == 16)

    @patch('torchbearer.variational.distributions.torch.normal')
    def test_rsample_number(self, normal):
        dist = SimpleNormal(1, 0)

        normal.side_effect = lambda mu, std: mu
        self.assertTrue(dist.rsample(sample_shape=torch.Size([2])).sum() == 2)

        normal.side_effect = lambda mu, std: std
        self.assertTrue(dist.rsample(sample_shape=torch.Size([2])).sum() == 4)

    def test_log_prob_tensor(self):
        mu = torch.ones(2, 2)
        logvar = (torch.zeros(2, 2) + 2).log()

        dist = SimpleNormal(mu, logvar)

        self.assertTrue(((dist.log_prob(torch.ones(2, 2)) + 1.2655).abs() < 0.0001).all())

    def test_log_prob_number(self):
        logvar = math.log(2)

        dist = SimpleNormal(1, logvar)

        self.assertTrue(((dist.log_prob(1) + 1.2655).abs() < 0.0001).all())


class TestSimpleUniform(unittest.TestCase):
    @patch('torchbearer.variational.distributions.torch.rand')
    def test_rsample_tensor(self, rand):
        low = torch.ones(2, 2)
        high = torch.ones(2, 2) + 1

        dist = SimpleUniform(low, high)

        rand.side_effect = lambda shape, dtype, device: torch.ones(shape) / 2
        self.assertTrue(((dist.rsample(sample_shape=torch.Size([2])) - 1.5).abs() < 0.0001).all())

    @patch('torchbearer.variational.distributions.torch.rand')
    def test_rsample_number(self, rand):
        dist = SimpleUniform(1, 2)

        rand.side_effect = lambda shape, dtype, device: torch.ones(shape) / 2
        self.assertTrue(((dist.rsample(sample_shape=torch.Size([2])) - 1.5).abs() < 0.0001).all())

    def test_log_prob_tensor(self):
        low = torch.ones(2, 2)
        high = torch.ones(2, 2) + 1

        dist = SimpleUniform(low, high)

        self.assertTrue((dist.log_prob(torch.ones(2, 2)).abs() < 0.0001).all())
        self.assertTrue((dist.log_prob(torch.ones(2, 2) + 2) == float('-inf')).all())

    def test_log_prob_number(self):
        dist = SimpleUniform(1, 2)

        self.assertTrue((dist.log_prob(1).abs() < 0.0001).all())
        self.assertTrue((dist.log_prob(3) == float('-inf')).all())


class TestSimpleExponential(unittest.TestCase):
    @patch('torchbearer.variational.distributions.broadcast_all')
    def test_rsample_tensor(self, broadcast_all):
        rate = (torch.ones(2, 2) / 2).log()

        def new_mock_tensor(shape):
            x = torch.ones(shape)
            x.exponential_ = Mock(return_value=x)
            return x

        rate.new = Mock(side_effect=new_mock_tensor)
        broadcast_all.return_value = (rate,)
        dist = SimpleExponential(rate)

        self.assertTrue(((dist.rsample(sample_shape=torch.Size([2])) - 2.0).abs() < 0.0001).all())

    @patch('torchbearer.variational.distributions.broadcast_all')
    def test_rsample_number(self, broadcast_all):
        rate = (torch.ones(1) / 2).log()

        def new_mock_tensor(shape):
            x = torch.ones(shape)
            x.exponential_ = Mock(return_value=x)
            return x

        rate.new = Mock(side_effect=new_mock_tensor)
        broadcast_all.return_value = (rate,)
        dist = SimpleExponential(0.5)

        self.assertTrue(((dist.rsample(sample_shape=torch.Size([2])) - 2.0).abs() < 0.0001).all())

    def test_log_prob_tensor(self):
        dist = SimpleExponential((torch.ones(2, 2) / 2).log())

        self.assertTrue(((dist.log_prob(torch.ones(2, 2)) + 1.1931).abs() < 0.0001).all())

    def test_log_prob_number(self):
        dist = SimpleExponential(math.log(0.5))

        self.assertTrue(((dist.log_prob(torch.ones(2, 2)) + 1.1931).abs() < 0.0001).all())


class TestSimpleWeibull(unittest.TestCase):
    @patch('torchbearer.variational.distributions.torch.rand')
    def test_rsample_tensor(self, rand):
        l = torch.ones(2, 2)
        k = torch.ones(2, 2)

        dist = SimpleWeibull(l, k)

        rand.side_effect = lambda shape, dtype, device: torch.ones(shape) / 2
        self.assertTrue(((dist.rsample(sample_shape=torch.Size([2])) - 0.6931).abs() < 0.0001).all())

    @patch('torchbearer.variational.distributions.torch.rand')
    def test_rsample_number(self, rand):
        dist = SimpleWeibull(1, 1)

        rand.side_effect = lambda shape, dtype, device: torch.ones(shape) / 2
        self.assertTrue(((dist.rsample(sample_shape=torch.Size([2])) - 0.6931).abs() < 0.0001).all())

    def test_log_prob_tensor(self):
        l = torch.ones(2, 2)
        k = torch.ones(2, 2)

        dist = SimpleWeibull(l, k)
        self.assertTrue((dist.log_prob(torch.ones(2, 2)) < 0.0001).all())
        self.assertTrue((dist.log_prob(torch.ones(2, 2) - 2) == float('-inf')).all())

    def test_log_prob_number(self):
        dist = SimpleWeibull(1, 1)

        self.assertTrue((dist.log_prob(1) < 0.0001).all())
        self.assertTrue((dist.log_prob(-1) == float('-inf')).all())




"""
Distributions
------------------------------------

..  automodule:: variational.distributions
        :members:
        :undoc-members:

Divergences
------------------------------------

..  automodule:: variational.divergence
        :members:
        :undoc-members:

Auto-Encoding
------------------------------------

..  automodule:: variational.auto_encoder
        :members:
        :undoc-members:

Datasets
------------------------------------

..  automodule:: variational.datasets
        :members:
        :undoc-members:

Visualisation
------------------------------------

..  automodule:: variational.visualisation
        :members:
        :undoc-members:
"""

from .version import __version__

from .auto_encoder import *
from .distributions import *
from .divergence import *
from .datasets import *
from .visualisation import *



import torch.nn.modules as nn

import torchbearer


class AutoEncoderBase(nn.Module):
    def __init__(self, latent_dims):
        super(AutoEncoderBase, self).__init__()
        self.latent_dims = latent_dims

    def encode(self, x, state=None):
        """Encode the given batch of images and return latent space sample for each.

        Args:
            x: Batch of images to encode
            state: The trial state

        Returns:
            Encoded samples / tuple of samples for different spaces
        """
        raise NotImplementedError

    def decode(self, sample, state=None):
        """Decode the given latent space sample batch to images.

        Args:
            sample: The latent space samples
            state: The trial state

        Returns:
            Decoded images
        """
        raise NotImplementedError

    def forward(self, x, state=None):
        """Encode then decode the inputs, returning the result. Also binds the target as the input images in state.

        Args:
            x: Model input batch
            state: The trial state

        Returns:
            Auto-Encoded images
        """
        if state is not None:
            state[torchbearer.Y_TRUE] = x

        return self.decode(self.encode(x, state), state)



import os
import shutil
import zipfile

import numpy as np
from PIL import Image
from torch.utils.data import Dataset


def make_dataset(dir, extensions):
    from torchvision.datasets.folder import has_file_allowed_extension
    images = []
    for root, _, fnames in sorted(os.walk(dir)):
        for fname in sorted(fnames):
            if has_file_allowed_extension(fname, extensions):
                path = os.path.join(root, fname)
                item = path
                images.append(item)
    return images


class SimpleImageFolder(Dataset):
    def __init__(self, root, loader=None, extensions=None, transform=None, target_transform=None):
        """
        Simple image folder dataset that loads all images from inside a folder and returns items in (image, image) tuple

        Args:
            root (str): Root directory of dataset containing all aligned images
            loader (function, optional): Image loader function that takes a file or path and returns the loaded image (see torchvision.datasets.folder)
            extensions (:obj:`list` of :obj:`str`, optional): List of file extensions that can be loaded
            transform (``Transform``, optional): A function/transform that takes in an PIL image and returns a transformed version. E.g, ``transforms.RandomCrop``
            target_transform (``Transform``, optional): A function/transform that takes in an PIL image and returns a transformed version. E.g, ``transforms.RandomCrop``
        """
        from torchvision.datasets.folder import default_loader, IMG_EXTENSIONS
        loader = default_loader if loader is None else loader
        extensions = IMG_EXTENSIONS if extensions is None else extensions

        samples = make_dataset(root, extensions)

        self.root = root
        self.loader = loader
        self.extensions = extensions

        self.samples = samples

        self.transform = transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        """
        Args:
            index (int): Index of image

        Returns:
            tuple: (sample, target) where target is target transformed image.
        """
        path = self.samples[index]
        sample = self.loader(path)
        input_sample, target_sample = sample, sample
        if self.transform is not None:
            input_sample = self.transform(sample)
        if self.target_transform is not None:
            target_sample = self.target_transform(sample)

        return input_sample, target_sample

    def __len__(self):
        return len(self.samples)


class CelebA(SimpleImageFolder):
    def __init__(self, root, transform=None, target_transform=None):
        """
        `CelebA <http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html>`_ auto-encoding dataset

        Args:
            root (str): Root directory of dataset containing all aligned images in 'root'
            transform (``Transform``, optional): A function/transform that takes in an PIL image and returns a transformed version. E.g, ``transforms.RandomCrop``
            target_transform (``Transform``, optional): A function/transform that takes in an PIL image and returns a transformed version. E.g, ``transforms.RandomCrop``
        """
        super(CelebA, self).__init__(root, transform=transform, target_transform=target_transform)

    def __getitem__(self, index):
        item = super(CelebA, self).__getitem__(index)
        return item


class CelebA_HQ(SimpleImageFolder):
    def __init__(self, root, as_npy=False, transform=None):
        """
        CelebA_HQ, high quality version of `celebA <http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html>`_ auto-encoding dataset as introduced by `Progressive GAN <https://arxiv.org/abs/1710.10196>`_

        Args:
            root (str): Root directory of dataset containing all hq images in 'root'
            as_npy (bool, optional): If True, assume images are stored in numpy arrays. Else assume a standard image format
            transform (``Transform``, optional): A function/transform that takes in an PIL image and returns a transformed version. E.g, ``transforms.RandomCrop``
        """
        from torchvision.datasets.folder import default_loader, IMG_EXTENSIONS
        if as_npy:
            loader = self.npy_loader
            extensions = ['npy']
        else:
            loader = default_loader
            extensions = IMG_EXTENSIONS
        super(CelebA_HQ, self).__init__(root, loader, extensions, transform)

    @staticmethod
    def npy_loader(path):
        img = np.load(path)[0].transpose([1,2,0])
        pil_image = Image.fromarray(img)
        return pil_image

    def __getitem__(self, index):
        item = super(CelebA_HQ, self).__getitem__(index)
        return item


class dSprites(Dataset):
    def __init__(self, root, download=False, transform=None):
        """
        `dSprites <https://github.com/deepmind/dsprites-dataset>`_ Dataset

        Args:
            root (str): Root directory of dataset containing 'dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz' or to download it to
            download (bool, optional): If true, downloads the dataset from the internet and puts it in root directory. If dataset is already downloaded, it is not downloaded again.
            transform (``Transform``, optional): A function/transform that takes in an PIL image and returns a transformed version. E.g, ``transforms.RandomCrop``
        """
        super(dSprites, self).__init__()
        self.file = root
        self.transform = transform

        if download:
            self.download()

        self.data = self.load_data()
        self.latents_sizes = np.array([1, 3, 6, 40, 32, 32])
        self.latents_bases = np.concatenate((self.latents_sizes[::-1].cumprod()[::-1][1:], np.array([1, ])))
        self.latents_values = np.load(os.path.join(self.file, "latents_values.npy"))
        self.latents_classes = np.load(os.path.join(self.file, "latents_classes.npy"))

    def download(self):
        if not os.path.exists(os.path.join(self.file, "imgs.npy")):
            data_url = 'https://github.com/deepmind/dsprites-dataset/blob/master/dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz?raw=true'
            import sys
            if sys.version_info[0] < 3:
                import urllib2 as request
            else:
                import urllib.request as request
            file = os.path.join(self.file, "dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz")

            os.makedirs(self.file, exist_ok=True)
            with request.urlopen(data_url) as response, open(file, 'wb+') as out_file:
                shutil.copyfileobj(response, out_file)

            zip_ref = zipfile.ZipFile(file, 'r')
            zip_ref.extractall(self.file)
            zip_ref.close()

    def get_img_by_latent(self, latent_code):
        """
        Returns the image defined by the latent code
        
        Args:
            latent_code (:obj:`list` of :obj:`int`): Latent code of length 6 defining each generative factor

        Returns:
            Image defined by given code

        """
        def latent_to_index(latents):
            return np.dot(latents, self.latents_bases).astype(int)
        idx = latent_to_index(latent_code)
        return self.__getitem__(idx)[0]

    def load_data(self):
        root = os.path.join(self.file, "imgs.npy")
        data = np.load(root)
        return data

    def __getitem__(self, index):
        data = self.data[index]
        data = Image.fromarray(data * 255, mode='L')

        if self.transform is not None:
            data = self.transform(data)

        return data, data

    def __len__(self):
        return self.data.shape[0]



import unittest
from mock import Mock, patch

import torchbearer.variational.datasets as ds


class TestMakeDataset(unittest.TestCase):
    @patch('os.walk')
    def test_make_dataset_all_files(self, mock_walk):
        mock_walk.return_value = [
            ('/root', (), ('b.a', 'c.a'))
        ]

        out = ds.make_dataset('a', ('a'))
        self.assertTrue(out == ['/root/b.a', '/root/c.a'])

    @patch('os.walk')
    def test_make_dataset_some_files(self, mock_walk):
        mock_walk.return_value = [
            ('/root', (), ('b.v', 'c.a'))
        ]

        out = ds.make_dataset('a', ('a'))
        self.assertTrue(out == ['/root/c.a'])

    @patch('os.walk')
    def test_make_dataset_no_files(self, mock_walk):
        mock_walk.return_value = [
            ('/root', (), ('b.v', 'c.v'))
        ]

        out = ds.make_dataset('a', ('a'))
        self.assertTrue(out == [])

    @patch('os.walk')
    def test_make_dataset_no_root(self, mock_walk):
        mock_walk.return_value = [
            ('', (), ('b.a', 'c.a'))
        ]

        out = ds.make_dataset('a', ('a'))
        self.assertTrue(out == ['b.a', 'c.a'])

    @patch('os.walk')
    def test_make_dataset_multiple_extensions(self, mock_walk):
        mock_walk.return_value = [
            ('', (), ('b.1', 'c.2'))
        ]

        out = ds.make_dataset('a', ('1','2'))
        self.assertTrue(out == ['b.1', 'c.2'])


class TestSimpleImageFolder(unittest.TestCase):
    @patch.object(ds, 'make_dataset')
    def test_length(self, mock_md):
        loader = Mock()
        extensions = None
        transform = Mock()
        target_transform = Mock()

        images = ['a.1', 'b.1', 'c.1', 'd.1']
        mock_md.return_value = images
        dataset = ds.SimpleImageFolder('root', loader, extensions, transform, target_transform)

        self.assertTrue(len(dataset) == len(images))

    @patch.object(ds, 'make_dataset')
    def test_image_path(self, mock_md):
        loader = Mock()
        extensions = None
        transform = Mock()
        target_transform = Mock()

        images = ['a.1', 'b.1', 'c.1', 'd.1']
        mock_md.return_value = images
        dataset = ds.SimpleImageFolder('root', loader, extensions, transform, target_transform)

        _ = dataset[1]
        self.assertTrue(loader.call_args[0][0] == images[1])

        _ = dataset[2]
        self.assertTrue(loader.call_args[0][0] == images[2])

    @patch.object(ds, 'make_dataset')
    def test_transform_args(self, mock_md):
        loader = Mock()
        mock_image = Mock()
        loader.return_value = mock_image
        extensions = None
        transform = Mock()
        target_transform = Mock()

        images = ['a.1', 'b.1', 'c.1', 'd.1']
        mock_md.return_value = images
        dataset = ds.SimpleImageFolder('root', loader, extensions, transform, target_transform)

        _ = dataset[1]
        self.assertTrue(transform.call_args[0][0] == mock_image)

    @patch.object(ds, 'make_dataset')
    def test_target_transform_args(self, mock_md):
        loader = Mock()
        mock_image = Mock()
        loader.return_value = mock_image
        extensions = None
        transform = Mock()
        target_transform = Mock()

        images = ['a.1', 'b.1', 'c.1', 'd.1']
        mock_md.return_value = images
        dataset = ds.SimpleImageFolder('root', loader, extensions, transform, target_transform)

        _ = dataset[1]
        self.assertTrue(target_transform.call_args[0][0] == mock_image)


class TestCelebA(unittest.TestCase):
    @patch.object(ds.SimpleImageFolder, '__len__')
    @patch.object(ds.SimpleImageFolder, '__getitem__')
    def test_celeba(self, mock_gi, mock_len):
        return_val = 'test'
        mock_len.return_value = 10
        mock_gi.return_value = return_val

        celeba = ds.CelebA('root', Mock(), Mock())
        out = celeba[6]
        self.assertTrue(mock_gi.call_args[0][0] == 6)
        self.assertTrue(out == return_val)

    @patch.object(ds.SimpleImageFolder, '__len__')
    @patch.object(ds.SimpleImageFolder, '__getitem__')
    def test_celeba_hq_get_item(self, mock_gi, mock_len):
        return_val = 'test'
        mock_len.return_value = 10
        mock_gi.return_value = return_val

        celebahq = ds.CelebA_HQ('root', False, Mock())
        out = celebahq[6]

        self.assertTrue(mock_gi.call_args[0][0] == 6)
        self.assertTrue(out == return_val)

    @patch.object(ds.CelebA_HQ, 'npy_loader')
    @patch.object(ds.SimpleImageFolder, '__len__')
    @patch.object(ds.SimpleImageFolder, '__getitem__')
    def test_celeba_hq_use_numpy_loader(self, mock_gi, mock_len, mock_loader):
        mock_len.return_value = 10
        mock_gi.return_value = Mock()

        celebahq = ds.CelebA_HQ('root', True, Mock())
        out = celebahq[6]

        self.assertTrue(celebahq.loader == mock_loader)
        self.assertTrue(celebahq.extensions == ['npy'])

    @patch.object(ds.SimpleImageFolder, '__len__')
    @patch.object(ds.SimpleImageFolder, '__getitem__')
    @patch('torchvision.datasets.folder.default_loader')
    def test_celeba_hq_use_default_loader(self, mock_loader, mock_gi, mock_len, ):
        import torchvision
        mock_len.return_value = 10

        celebahq = ds.CelebA_HQ('root', False, Mock())
        out = celebahq[6]

        self.assertTrue(celebahq.loader == mock_loader)
        self.assertTrue(celebahq.extensions == torchvision.datasets.folder.IMG_EXTENSIONS)

    @patch.object(ds.SimpleImageFolder, '__len__')
    @patch('numpy.load')
    @patch('PIL.Image.fromarray')
    def test_celeba_hq_numpy_loader(self, mock_fromarray, mock_load, mock_len):
        samples = ['a.npy', 'b.npy', 'c.npy']
        mock_len.return_value = 3
        mock_output = Mock()
        mock_output.transpose.return_value = mock_output
        mock_load.return_value = (mock_output, )

        celebahq = ds.CelebA_HQ('root', True, Mock())
        celebahq.samples = samples
        out = celebahq[1]

        self.assertTrue(mock_load.call_args[0][0] == samples[1])
        self.assertTrue(mock_fromarray.call_args[0][0] == mock_output)


class TestdSprites(unittest.TestCase):
    @patch.object(ds.dSprites, 'download')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_no_download(self, mock_load, mock_load_data, mock_download):
        sprites = ds.dSprites('root', False, None)
        self.assertTrue(sprites.download.call_count == 0)

    @patch.object(ds.dSprites, 'download')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_do_download(self, mock_load, mock_load_data, mock_download):
        sprites = ds.dSprites('root', True, None)
        self.assertTrue(sprites.download.call_count == 1)

    @patch.object(ds.dSprites, 'download')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_latents_bases(self, mock_load, mock_load_data, mock_download):
        sprites = ds.dSprites('root', False, None)
        self.assertTrue(list(sprites.latents_bases) == [737280, 245760, 40960, 1024, 32, 1])

    @patch.object(ds.dSprites, 'download')
    @patch('numpy.load')
    def test_load_data(self, mock_load, mock_download):
        sprites = ds.dSprites('root', False, None)
        self.assertTrue(mock_load.call_args_list[0][0][0] == 'root/imgs.npy')
        self.assertTrue(mock_load.call_args_list[1][0][0] == 'root/latents_values.npy')
        self.assertTrue(mock_load.call_args_list[2][0][0] == 'root/latents_classes.npy')

    @patch.object(ds.dSprites, 'download')
    @patch.object(ds.dSprites, '__getitem__')
    @patch('numpy.load')
    def test_image_by_latent(self, mock_load, mock_gi, mock_download):
        import numpy as np
        mock_gi.return_value = (None, )

        bases = np.array([737280, 245760, 40960, 1024, 32, 1])
        latent = [1,2,3,4,5,6]
        img_id = (bases * latent).sum()

        sprites = ds.dSprites('root', False, None)
        sprites.get_img_by_latent(latent)
        self.assertTrue(sprites.__getitem__.call_args[0][0] == img_id)

    @patch.object(ds.dSprites, 'download')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_len(self, mock_load, mock_load_data, mock_download):
        import numpy as np
        mock_load_data.return_value = np.zeros((10, 5, 5))
        sprites = ds.dSprites('root', False, None)

        self.assertTrue(len(sprites) == 10)

    @patch('PIL.Image.fromarray')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_get_item(self, mock_load, mock_load_data, mock_fromarray):
        import torch
        sprites = ds.dSprites('root', False, None)
        mock_data = torch.rand(5, 1)
        mock_data[1] *= 3
        sprites.data = mock_data
        out = sprites[1]

        self.assertTrue(mock_fromarray.call_args[0][0] == mock_data[1]*255)
        self.assertTrue(mock_fromarray.call_args[1] == {'mode':'L'})
        self.assertTrue(len(out) == 2)

    @patch('PIL.Image.fromarray')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_get_item_transform(self, mock_load, mock_load_data, mock_fromarray):
        import torch
        transform = Mock()
        sprites = ds.dSprites('root', False, transform)
        mock_fromarray.return_value = 'test'
        mock_data = torch.rand(5, 1)
        mock_data[1] *= 3
        sprites.data = mock_data
        _ = sprites[1]

        self.assertTrue(transform.call_args[0][0] == 'test')

    @patch('zipfile.ZipFile')
    @patch('shutil.copyfileobj')
    @patch('os.makedirs')
    @patch.object(ds.dSprites, 'load_data')
    @patch('numpy.load')
    def test_download(self, mock_load, mock_load_data, mock_mkdirs, mock_copyfileobj, mock_zip, mock_open, mock_urlopen):
        mock_return = Mock()
        # mock_urlopen = Mock()
        # mock_url.return_value = mock_urlopen
        mock_urlopen.return_value.__enter__.return_value = mock_return

        root = 'root'
        filename = "dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz"

        sprites = ds.dSprites(root, True, None)

        self.assertTrue(mock_mkdirs.call_args[0][0] == "root")
        self.assertTrue(mock_mkdirs.call_args[1] == {'exist_ok': True})

        self.assertTrue(mock_urlopen.call_args[0][0] == 'https://github.com/deepmind/dsprites-dataset/blob/master/dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz?raw=true')
        self.assertTrue(mock_open.call_args[0][0] == 'root/' + filename)

        self.assertTrue(mock_copyfileobj.call_args[0][0] == mock_return)

    import sys
    if sys.version_info[0] < 3:
        test_download = patch('urllib2.urlopen')(patch('__builtin__.open')(test_download))
    else:
        test_download = patch('urllib.request.urlopen')(patch('builtins.open')(test_download))


import torch

import torchbearer
import torchbearer.callbacks as c


class LatentWalker(c.Callback):
    def __init__(self, same_image, row_size):
        """
        Args:
            same_image (bool): If True, use the same image for all latent dimension walks. Else each dimension has different image
            row_size (int): Number of images displayed in each row of the grid.
        """
        super(LatentWalker, self).__init__()
        self.data_key = None
        self.same_image = same_image
        self.row_size = row_size
        self.model = None
        self.data = None
        self.dev = None
        self.file = None
        self.store_key = None
        self.variable_space = 0

    def on_train(self):
        """
        Sets the walker to run during training

        Returns:
            LatentWalker: self
        """
        self.on_step_training = c.once_per_epoch(self._vis)
        return self

    def on_val(self):
        """
        Sets the walker to run during validation

        Returns:
            LatentWalker: self
        """
        self.on_step_validation = c.once_per_epoch(self._vis)
        return self

    def for_space(self, space_id):
        """
        Sets the ID for which latent space to vary when model outputs [latent_space_0, latent_space_1, ...]

        Args:
            space_id (int): ID of the latent space to vary

        Returns:
            LatentWalker: self
        """
        self.variable_space = space_id
        return self

    def for_data(self, data_key):
        """
        Args:
            data_key (:class:`.StateKey`): State key which will contain data to act on

        Returns:
            LatentWalker: self
        """
        self.data_key = data_key
        return self

    def to_key(self, state_key):
        """

        Args:
            state_key (:class:`.StateKey`): State key under which to store result

        Returns:
            LatentWalker: self
        """
        self.store_key = state_key
        return self

    def to_file(self, file):
        """
        Args:
            file (string, pathlib.Path object or file object): File in which result is saved

        Returns:
            LatentWalker: self
        """
        self.file = file
        return self

    def _vis(self, state):
        self.model = state[torchbearer.MODEL]
        self.data = state[self.data_key] if self.data_key is not None else state[torchbearer.X]
        self.dev = state[torchbearer.DEVICE]

        with torch.no_grad():
            result = self.vis(state)

        if self.file is not None:
            self._save_walk(result)
        if self.store_key is not None:
            state[self.store_key] = result

    def vis(self, state):
        """
        Create the tensor of images to be displayed
        """
        raise NotImplementedError

    def _save_walk(self, tensor):
        from torchvision.utils import save_image
        save_image(tensor, self.file, self.row_size, normalize=True, pad_value=1)


class ReconstructionViewer(LatentWalker):
    def __init__(self, row_size=8, recon_key=torchbearer.Y_PRED):
        """
        Latent space walker that just returns the reconstructed images for the batch

        Args:
            row_size (int): Number of images displayed in each row of the grid.
            recon_key (StateKey): :class:`.StateKey` of the reconstructed images
        """
        super(ReconstructionViewer, self).__init__(False, row_size)
        self.recon_key = recon_key

    def vis(self, state):
        data = self.data[:self.row_size]
        recons = state[self.recon_key][:self.row_size]
        return torch.cat([data, recons])


class LinSpaceWalker(LatentWalker):
    def __init__(self, lin_start=-1, lin_end=1, lin_steps=8, dims_to_walk=[0], zero_init=False, same_image=False):
        """
        Latent space walker that explores each dimension linearly from start to end points

        Args:
            lin_start (float): Starting point of linspace
            lin_end (float): End point of linspace
            lin_steps (int): Number of steps to take in linspace
            dims_to_walk (list of int): List of dimensions to walk
            zero_init (bool): If True, dimensions not being walked are 0. Else, they are obtained from encoder
            same_image (bool): If True, use same image for each dimension walked. Else, use different images
        """
        super(LinSpaceWalker, self).__init__(same_image, lin_steps)
        self.dims_to_walk = dims_to_walk
        self.zero_init = zero_init
        self.linspace = torch.linspace(lin_start, lin_end, lin_steps)

    def vis(self, state):
        self.linspace = self.linspace.to(self.dev)
        num_images = self.row_size * len(self.dims_to_walk)
        num_spaces = len(self.model.latent_dims)

        if self.zero_init:
            sample = []
            for i in range(num_spaces):
                sample.append(torch.zeros(num_images, self.model.latent_dims[i], device=self.dev).unsqueeze(1).repeat(1, self.row_size, 1))
        elif self.same_image:
            sample = list(self.model.encode(self.data[0], state))
            for i in range(len(sample)):
                sample[i] = sample[i][0].unsqueeze(0).unsqueeze(1).repeat(sample[i].shape[0], self.row_size, 1)
        else:
            sample = list(self.model.encode(self.data, state))
            for i in range(len(sample)):
                sample[i] = sample[i].unsqueeze(1).repeat(1, self.row_size, 1)

        dims = self.dims_to_walk

        i = 0
        for dim in list(dims):
            sample[self.variable_space][i, :, dim] = self.linspace
            i += 1

        for i in range(num_spaces):
            sample[i] = sample[i].view(-1, self.model.latent_dims[i])[:num_images]

        result = self.model.decode(sample).view(num_images, -1, self.data.shape[-2], self.data.shape[-1])
        return result


class RandomWalker(LatentWalker):
    def __init__(self, var=1, num_images=32, uniform=False, row_size=8):
        """
        Latent space walker that shows random samples from latent space

        Args:
            var (float or torch.Tensor): Variance of random sample
            num_images (int): Number of random images to sample
            uniform (bool): If True, sample uniform distribution [-v, v). If False, sample normal distribution with var v
            row_size (int): Number of images displayed in each row of the grid.
        """
        super(RandomWalker, self).__init__(False, row_size)
        self.num_images = num_images
        self.uniform = uniform
        self.var = var

    def vis(self, state):
        num_spaces = len(self.model.latent_dims)

        sample = []
        for i in range(num_spaces):
            sample.append(torch.zeros(self.num_images, self.model.latent_dims[i], device=self.dev))

        if self.uniform:
            sample[self.variable_space] = (torch.rand(self.num_images, self.model.latent_dims[self.variable_space], device=self.dev)*2-1)*self.var
        else:
            sample[self.variable_space] = (torch.randn(self.num_images, self.model.latent_dims[self.variable_space], device=self.dev)*2-1)*self.var

        result = self.model.decode(sample).view(sample[0].shape[0], -1, self.data.shape[-2], self.data.shape[-1])
        return result


class CodePathWalker(LatentWalker):
    def __init__(self, num_steps, p1, p2):
        """
        Latent space walker that walks between two specified codes p1 and p2

        Args:
            num_steps (int): Number of steps to take between points
            p1 (torch.Tensor): Batch of codes
            p2 (torch.Tensor): Batch of codes
        """
        super(CodePathWalker, self).__init__(True, num_steps)
        self.p1 = p1
        self.p2 = p2
        self.num_steps = num_steps

    def vis(self, state):
        step_sizes = (self.p1 - self.p2)/(self.num_steps-1)

        codes = torch.zeros(self.p1.shape[0], self.num_steps, self.p1.shape[1]).to(self.dev)
        for i in range(self.num_steps):
            codes[:, i] = self.p1 - step_sizes*i
        codes = codes.view(-1, self.p1.shape[1])

        result = self.model.decode(codes).view(codes.shape[0], -1, self.data.shape[-2], self.data.shape[-1])
        return result


class ImagePathWalker(CodePathWalker):
    def __init__(self, num_steps, im1, im2):
        """
        Latent space walker that walks between two specified images im1 and im2

        Args:
            num_steps (int): Number of steps to take between points
            im1 (torch.Tensor): Batch of images
            im2 (torch.Tensor): Batch of images
        """
        super(ImagePathWalker, self).__init__(num_steps, None, None)
        self.im1, self.im2 = im1, im2

    def vis(self, state):
        self.p1 = self.model.encode(self.im1.to(self.dev), state)
        self.p2 = self.model.encode(self.im2.to(self.dev), state)

        return super(ImagePathWalker, self).vis(state)
